/*
 * Copyright (c) 2015, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * WSO2 Inc. licenses this file to you under the Apache License,
 * Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.wso2.andes.kernel.disruptor.inbound;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.andes.kernel.AndesContextInformationManager;
import org.wso2.andes.kernel.AndesException;

/**
 * Exchange related inbound event 
 */
public class InboundExchangeEvent implements AndesInboundStateEvent {

    private static Log log = LogFactory.getLog(InboundExchangeEvent.class);

    private enum EventType {

        CREATE_EXCHANGE_EVENT,

        DELETE_EXCHANGE_EVENT
    }

    /**
     * Name of the exchange
     */
    private String messageRouterName;

    /**
     * Type of the exchange
     */
    private String type;

    /**
     * Is exchange deletable
     */

    private boolean autoDelete;

    /**
     * Event type this event
     */
    private EventType eventType;

    /**
     * Reference to AndesContextInformationManager to update create/ remove queue state
     */
    private AndesContextInformationManager contextInformationManager;

    /**
     * create an instance of Andes Exchange
     *
     * @param messageRouterName name of exchange
     * @param type         type of the exchange (use to pass additional information
     *                     regarding an exchange)
     * @param autoDelete   is exchange auto deletable (
     *                     if true exchange is removed when all queues are detached)
     */
    public InboundExchangeEvent(String messageRouterName, String type, boolean autoDelete) {
       this.messageRouterName = messageRouterName;
       this.type = type;
       this.autoDelete = autoDelete;
    }

    public boolean isAutoDelete() {
        return autoDelete;
    }

    public String getType() {
        return type;
    }

    public String getMessageRouterName() {
        return messageRouterName;
    }

    @Override
    public void updateState() throws AndesException {
        switch (eventType) {
            case CREATE_EXCHANGE_EVENT:
                contextInformationManager.createExchange(this);
                break;
            case DELETE_EXCHANGE_EVENT:
                contextInformationManager.deleteExchange(this);
                break;
            default:
                log.error("Event type not set properly " + eventType);
                break;
        }
    }

    @Override
    public String eventInfo() {
        return eventType.toString();
    }

    @Override
    public boolean isActionableWhenPassive() {
        return true;
    }

    /**
     * Update the event to a create exchange event
     * @param contextInformationManager AndesContextInformationManager
     */
    public void prepareForCreateExchange(AndesContextInformationManager contextInformationManager) {
        eventType = EventType.CREATE_EXCHANGE_EVENT;
        this.contextInformationManager = contextInformationManager;
    }

    /**
     * Update event to delete exchange event
     * @param contextInformationManager AndesContextInformationManager
     */
    public void prepareForDeleteExchange(AndesContextInformationManager contextInformationManager) {
        eventType = EventType.DELETE_EXCHANGE_EVENT;
        this.contextInformationManager = contextInformationManager;
    }
}
